//: SSHAPE2.CPP -- Alternative to SSHAPE.CPP
#include <iostream.h>

class shape {
  shape(shape&); // Prevent copy-construction
protected:
  shape() {} // Prevent stack objects
  // But allow access to derived constructors
public:
  enum type { Circle, Square, Triangle };
  virtual void draw() = 0;
  virtual ~shape() { cout << "~shape\n"; }
  static shape* make(type);
};

class circle : public shape {
  circle(circle&); // No copy-construction
  circle operator=(circle&); // No operator=
protected:
  circle() {};
public:
  void draw() { cout << "circle::draw\n"; }
  ~circle() { cout << "~circle\n"; }
  friend shape* shape::make(type t);
};

class square : public shape {
  square(square&); // No copy-construction
  square operator=(square&); // No operator=
protected:
  square() {};
public:
  void draw() { cout << "square::draw\n"; }
  ~square() { cout << "~square\n"; }
  friend shape* shape::make(type t);
};

class triangle : public shape {
  triangle(triangle&); // No copy-construction
  triangle operator=(triangle&); // Prevent
protected:
  triangle() {};
public:
  void draw() { cout << "triangle::draw\n"; }
  ~triangle() { cout << "~triangle\n"; }
  friend shape* shape::make(type t);
};

shape* shape::make(type t) {
  shape* S;
  switch(t) {
    case Circle: S = new circle; break;
    case Square: S = new square; break;
    case Triangle: S = new triangle; break;
  }
  S->draw(); // Virtual function call
  return S;
}

#include <vector.h>
typedef vector<shape*> container;
typedef container::iterator iter;

int main() {
  container shapes; // Default to ownership
  shapes.push_back(shape::make(shape::Circle));
  shapes.push_back(shape::make(shape::Square));
  shapes.push_back(shape::make(shape::Triangle));
  cout << "virtual function calls:\n";
  for(iter i = shapes.begin(); i != shapes.end(); i++)
    (*i)->draw();
  //Circle c; // error: can't create on stack
  for(iter j = shapes.begin(); j != shapes.end(); j++)
    delete *j;
  return 0;
}
